import daj.Program;
import java.util.*;

//
//
// MainThread
//
//
class MainThread extends InternalThread
{
	/**
		The interval between checks for the state of system
	*/
	public static final int CHECK_INTERVAL = 5;

	/**
		The constructor of the main thread
	*/
	public MainThread (Processor processor)
	{
		super (processor);
	}

	/**
		The program executed on this thread
	*/
	public void main ()
	{
		Random random = new Random ();

		while (true)
		{
			sleep (CHECK_INTERVAL);
			if (processor.State == Processor.NORMAL && processor.Coordinator == processor.index)
			{
				if (Math.abs (random.nextInt ()) % 2 == 0)
				{
					Check ();
					continue;
				}
			}
			Timeout ();
		}
	}

	/**
		Searches for other groups to merge with
	*/
	protected void Check ()
	{
		BitSet Others = new BitSet (processor.getNeighborsCount ());
		send (new AreYouCoordMsg ());

		int time = getTime ();
		boolean empty = true;
		while (getTime () - time < Processor.TIMEOUT)
		{
			InternalMessage msg = receive ();
			if (msg != null && msg.outerMessage instanceof AreYouCoordAnswer)
			{
				AreYouCoordAnswer answer = (AreYouCoordAnswer)msg.outerMessage;
				if (answer.isCoordinator)
				{
					Others.set (msg.channel);
					empty = false;
				}
			}
		}
		if (empty)
			return;
		sleep (1 + Processor.DELAY * 5);
		Merge (Others);
	}

	/**
		Tests the failure of the coordinator
	*/
	protected void Timeout ()
	{
		if (processor.Coordinator == processor.index)
			return;
		send (processor.Coordinator, new AreYouThereMsg (processor.Group));
		InternalMessage msg = receive ();
		if (msg != null && msg.outerMessage instanceof AreYouThereAnswer)
		{
			if (((AreYouThereAnswer)msg.outerMessage).stillCoordinator)
				return;
		}
		processor.Recovery ();
	}

	/**
		Makes attempts to merge all the known groups into one
	*/
	public void Merge (BitSet Coordinators)
	{
		processor.State = Processor.ELECTION;
		processor.Group = new Group (processor.index, ++processor.Counter);
		processor.Coordinator = processor.index;
		BitSet upSet = processor.Up;
		processor.Up = new BitSet ();

		for (int i = 0;i < Coordinators.size ();i++)
			if (Coordinators.get (i))
				send (i, new InvitationMsg (processor.index, processor.Group));

		for (int i = 0;i < upSet.size ();i++)
			if (upSet.get (i))
				send (i, new InvitationMsg (processor.index, processor.Group));

		sleep (Processor.TIMEOUT * 5);

		processor.State = Processor.ELECTION;
		int numAnswers = 0;
		int UpSize = 0;
		for (int i = 0;i < processor.Up.size ();i++)
		{
			if (processor.Up.get (i))
			{
				send (i, new ReadyMsg (processor.Group));
				UpSize++;
			}
		}

		int time = getTime ();
		while (getTime () - time < Processor.TIMEOUT)
		{
			InternalMessage msg = receive ();
			if (msg != null && msg.outerMessage instanceof ReadyAnswer)
			{
				ReadyAnswer answer = (ReadyAnswer)msg.outerMessage;
				if (answer.inGroup && processor.Group.equals (answer.group))
					numAnswers++;
			}
		}
		if (numAnswers < UpSize)
			processor.Recovery ();
		else
			processor.State = Processor.NORMAL;
	}
}

